/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
package org.apache.woden.internal.util.dom;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.List;
import java.util.ListIterator;
import java.util.Vector;

import javax.xml.namespace.QName;

import org.apache.woden.ErrorReporter;
import org.apache.woden.WSDLException;
import org.apache.woden.internal.ErrorLocatorImpl;
import org.apache.woden.internal.ErrorReporterImpl;
import org.apache.woden.wsdl20.xml.DescriptionElement;
import org.w3c.dom.Attr;
import org.w3c.dom.CharacterData;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

/**
 * This class originated from WSDL4J.
 * 
 * @author jkaputin@apache.org (Woden changes)
 */
public class DOMUtils {
  /**
   * The namespaceURI represented by the prefix <code>xmlns</code>.
   */
  private static String NS_URI_XMLNS = "http://www.w3.org/2000/xmlns/";

  private static final String ATTR_XMLNS = "xmlns";
  private static final String emptyString = "";
  
  /**
   * Returns a list of attributes of an element. Returns an 
   * empty list if the element has no attributes. Does not
   * include namespace declarations.
   *
   * @param el       Element whose attributes are returned
   * @return the List of Attr
   */
  static public List getAttributes (Element el) {
    String nodename, prefix = null;
    List attrs = new Vector();
    NamedNodeMap attrMap = el.getAttributes();
    for(int i = 0; i < attrMap.getLength(); i++)
    {
      nodename = attrMap.item(i).getNodeName();
      prefix = attrMap.item(i).getPrefix();
      
      if (ATTR_XMLNS.equals(nodename) || ATTR_XMLNS.equals(prefix))
      {
        //ignore namespace declarations
        continue;
      }
      else
      {
        attrs.add(attrMap.item(i));  
      }
    }
      
    return attrs;
  }

  /**
   * Returns the value of an attribute of an element. Returns null
   * if the attribute is not found (whereas Element.getAttribute
   * returns "" if an attrib is not found). This method should be
   * used for elements that support extension attributes because it
   * does not track unexpected attributes.
   *
   * @param el       Element whose attrib is looked for
   * @param attrName name of attribute to look for
   * @return the attribute value including prefix if present
   */
  static public String getAttribute (Element el, String attrName) {
    String sRet = null;
    Attr   attr = el.getAttributeNode(attrName);

    if (attr != null) {
      sRet = attr.getValue();
    }
    return sRet;
  }

  /**
   * Returns the value of an attribute of an element. Returns null
   * if the attribute is not found (whereas Element.getAttribute
   * returns "" if an attrib is not found). This method should be
   * used for elements that do not support extension attributes
   * because it tracks the element's remaining attributes so that
   * eventually any unexpected attributes can be identified.
   *
   * @param el       Element whose attrib is looked for
   * @param attrName name of attribute to look for
   * @param remainingAttrs List of remaining attributes 
   * @return the attribute value
   */
  static public String getAttribute (Element el, String attrName, List remainingAttrs) {
    String sRet = null;
    Attr   attr = el.getAttributeNode(attrName);
    
    if (attr != null) {
      sRet = attr.getValue();
      remainingAttrs.remove(attr);
    }
    return sRet;
  }

  /**
   * Returns the value of an attribute of an element. Returns null
   * if the attribute is not found (whereas Element.getAttributeNS
   * returns "" if an attrib is not found).
   *
   * @param el       Element whose attrib is looked for
   * @param namespaceURI namespace URI of attribute to look for
   * @param localPart local part of attribute to look for
   * @return the attribute value
   */
  static public String getAttributeNS (Element el,
                                       String namespaceURI,
                                       String localPart) {
    String sRet = null;
    Attr   attr = el.getAttributeNodeNS (namespaceURI, localPart);

    if (attr != null) {
      sRet = attr.getValue ();
    }

    return sRet;
  }

  /**
   * Concat all the text and cdata node children of this elem and return
   * the resulting text.
   *
   * @param parentEl the element whose cdata/text node values are to
   *                 be combined.
   * @return the concatanated string.
   */
  static public String getChildCharacterData (Element parentEl) {
    if (parentEl == null) {
      return null;
    }
    Node          tempNode = parentEl.getFirstChild();
    StringBuffer  strBuf   = new StringBuffer();
    CharacterData charData;

    while (tempNode != null) {
      switch (tempNode.getNodeType()) {
        case Node.TEXT_NODE :
        case Node.CDATA_SECTION_NODE : charData = (CharacterData)tempNode;
                                       strBuf.append(charData.getData());
                                       break;
      }
      tempNode = tempNode.getNextSibling();
    }
    return strBuf.toString();
  }

  /**
   * Return the first child element of the given element. Null if no
   * children are found.
   *
   * @param elem Element whose child is to be returned
   * @return the first child element.
   */
  public static Element getFirstChildElement (Element elem) {
    for (Node n = elem.getFirstChild (); n != null; n = n.getNextSibling ()) {
      if (n.getNodeType () == Node.ELEMENT_NODE) {
        return (Element) n;
      }
    }
    return null;
  }

  /**
   * Return the next sibling element of the given element. Null if no
   * more sibling elements are found.
   *
   * @param elem Element whose sibling element is to be returned
   * @return the next sibling element.
   */
  public static Element getNextSiblingElement (Element elem) {
    for (Node n = elem.getNextSibling (); n != null; n = n.getNextSibling ()) {
      if (n.getNodeType () == Node.ELEMENT_NODE) {
        return (Element) n;
      }
    }
    return null;
  }

  /**
   * Return the first child element of the given element which has the
   * given attribute with the given value.
   *
   * @param elem      the element whose children are to be searched
   * @param attrName  the attrib that must be present
   * @param attrValue the desired value of the attribute
   *
   * @return the first matching child element.
   */
  public static Element findChildElementWithAttribute (Element elem,
                   String attrName,
                   String attrValue) {
    for (Node n = elem.getFirstChild (); n != null; n = n.getNextSibling ()) {
      if (n.getNodeType () == Node.ELEMENT_NODE) {
        if (attrValue.equals (DOMUtils.getAttribute ((Element) n, attrName))) {
          return (Element) n;
        }
      }
    }
    return  null;
  }

  /**
   * Count number of children of a certain type of the given element.
   *
   * @param elem the element whose kids are to be counted
   *
   * @return the number of matching kids.
   */
  public static int countKids (Element elem, short nodeType) {
    int nkids = 0;
    for (Node n = elem.getFirstChild (); n != null; n = n.getNextSibling ()) {
      if (n.getNodeType () == nodeType) {
        nkids++;
      }
    }
    return nkids;
  }

  /**
   * Given a prefix and a node, return the namespace URI that the prefix
   * has been associated with. This method is useful in resolving the
   * namespace URI of attribute values which are being interpreted as
   * QNames. If prefix is null, this method will return the default
   * namespace.
   *
   * @param context the starting node (looks up recursively from here)
   * @param prefix the prefix to find an xmlns:prefix=uri for
   *
   * @return the namespace URI or null if not found
   */
  public static String getNamespaceURIFromPrefix (Node context,
                                                  String prefix) {
    short nodeType = context.getNodeType ();
    Node tempNode = null;

    switch (nodeType)
    {
      case Node.ATTRIBUTE_NODE :
      {
        tempNode = ((Attr) context).getOwnerElement ();
        break;
      }
      case Node.ELEMENT_NODE :
      {
        tempNode = context;
        break;
      }
      default :
      {
        tempNode = context.getParentNode ();
        break;
      }
    }

    while (tempNode != null && tempNode.getNodeType () == Node.ELEMENT_NODE)
    {
      Element tempEl = (Element) tempNode;
      String namespaceURI = (prefix == null)
                            ? getAttribute (tempEl, ATTR_XMLNS)
                            : getAttributeNS (tempEl, NS_URI_XMLNS, prefix);

      if (namespaceURI != null)
      {
        return namespaceURI;
      }
      else
      {
        tempNode = tempEl.getParentNode ();
      }
    }

    return null;
  }
  
  /*
   * this is a new version of the original getQName method from WSDL4J, with the invocation
   * of registerUniquePrefix commented out and the 'desc' arg removed from the arg list.
   * TODO pending modification to original method post-M2 (see todo comments below).
   */
  public static QName getQName(String prefixedValue,
                               Element contextEl)
                               throws WSDLException
  {
      int    index        = prefixedValue.indexOf(':');
      String prefix       = (index != -1)
      ? prefixedValue.substring(0, index)
              : null;
      String localPart    = prefixedValue.substring(index + 1);
      String namespaceURI = getNamespaceURIFromPrefix(contextEl, prefix);
      
      if (namespaceURI != null)
      {
          //TODO investigate changing the registration of namespaces and prefixes (i.e. namespace decls)
          //so it can happen within any WSDL element, not just Description (current behaviour is copied from WSDL4J)
          //registerUniquePrefix(prefix, namespaceURI, desc);
          
          //TODO when passing prefix to QName ctor, what if it was modified by
          //registerUniquePrefix because of a name clash (pass original or modification)?
          return new QName(namespaceURI, 
                  localPart, 
                  prefix != null ? prefix : emptyString);
      }
      else
      {
          //TODO use ErrorReporter here or in callers to report the problem
          
          String faultCode = (prefix == null)
          ? WSDLException.NO_PREFIX_SPECIFIED
                  : WSDLException.UNBOUND_PREFIX;
          
          WSDLException wsdlExc = new WSDLException(faultCode,
                  "Unable to determine " +
                  "namespace of '" +
                  prefixedValue + "'.");
          
          wsdlExc.setLocation(XPathUtils.getXPathExprFromNode(contextEl));
          
          throw wsdlExc;
      }
  }

  /*
   * this is the original version of the getQName method, per WSDL4J. See to do comment below
   * about changing register unique prefix behaviour. TODO TBD post-M2 (JK)
   */
  public static QName getQName(String prefixedValue,
                               Element contextEl,
                               DescriptionElement desc)
                                 throws WSDLException
  {
    int    index        = prefixedValue.indexOf(':');
    String prefix       = (index != -1)
                          ? prefixedValue.substring(0, index)
                          : null;
    String localPart    = prefixedValue.substring(index + 1);
    String namespaceURI = getNamespaceURIFromPrefix(contextEl, prefix);

    if (namespaceURI != null)
    {
      //TODO investigate changing the registration of namespaces and prefixes (i.e. namespace decls)
      //so it can happen within any WSDL element, not just Description (current behaviour is copied from WSDL4J)
      registerUniquePrefix(prefix, namespaceURI, desc);
      
      //TODO when passing prefix to QName ctor, what if it was modified by
      //registerUniquePrefix because of a name clash (pass original or modification)?
      return new QName(namespaceURI, 
                       localPart, 
                       prefix != null ? prefix : emptyString);
    }
    else
    {
      //TODO use ErrorReporter here or in callers to report the problem
        
      String faultCode = (prefix == null)
                         ? WSDLException.NO_PREFIX_SPECIFIED
                         : WSDLException.UNBOUND_PREFIX;

      WSDLException wsdlExc = new WSDLException(faultCode,
                                                "Unable to determine " +
                                                "namespace of '" +
                                                prefixedValue + "'.");

      wsdlExc.setLocation(XPathUtils.getXPathExprFromNode(contextEl));

      throw wsdlExc;
    }
  }
  
  public static void registerUniquePrefix(String prefix,
                                          String namespaceURI,
                                          DescriptionElement desc)
  throws WSDLException
  {
    URI nsUri = desc.getNamespace(prefix);
    String tempNSUri = nsUri != null ? nsUri.toString() : null;

    if (tempNSUri != null && tempNSUri.equals(namespaceURI))
    {
      return;
    }

    while (tempNSUri != null && !tempNSUri.equals(namespaceURI))
    {
      prefix += "_";
      nsUri = desc.getNamespace(prefix);
      tempNSUri = nsUri != null ? nsUri.toString() : null;
    }
    
    URI uri = null;
    try {
        uri = new URI(namespaceURI);
    } catch (URISyntaxException e) {
        //TODO make sure custom err handler is used, if configured
        new ErrorReporterImpl().reportError(
                new ErrorLocatorImpl(),  //TODO line&col nos.
                "WSDL506", 
                new Object[] {namespaceURI}, 
                ErrorReporter.SEVERITY_ERROR, 
                e);
    }

    desc.addNamespace(prefix, uri);
  }

  /**
   * This method should be used for elements that support extension attributes
   * because it does not track the remaining attributes to test for unexpected 
   * attributes.
   */
  /*
  public static QName getQualifiedAttributeValue(Element el,
                                                 String attrName,
                                                 String elDesc,
                                                 boolean isRequired,
                                                 Definition def)
                                                   throws WSDLException
  {
    String attrValue = DOMUtils.getAttribute(el, attrName);

    if (attrValue != null)
    {
      return getQName(attrValue, el, def);
    }
    else if (isRequired)
    {
      WSDLException wsdlExc = new WSDLException(WSDLException.INVALID_WSDL,
                                                "The '" + attrName +
                                                "' attribute must be " +
                                                "specified for every " +
                                                elDesc + " element.");

      wsdlExc.setLocation(XPathUtils.getXPathExprFromNode(el));

      throw wsdlExc;
    }
    else
    {
      return null;
    }
  }
  /*

  /**
   * This method should be used for elements that do not support extension attributes
   * because it tracks the remaining attributes so that eventually any
   * unexpected attributes can be identified.
   */
  /*
  public static QName getQualifiedAttributeValue(Element el,
                                                 String attrName,
                                                 String elDesc,
                                                 boolean isRequired,
                                                 Definition def,
                                                 List remainingAttrs)
                                                   throws WSDLException
  {
    String attrValue = null;
    
    attrValue = DOMUtils.getAttribute(el, attrName, remainingAttrs);

    if (attrValue != null)
    {
      return getQName(attrValue, el, def);
    }
    else if (isRequired)
    {
      WSDLException wsdlExc = new WSDLException(WSDLException.INVALID_WSDL,
                                                "The '" + attrName +
                                                "' attribute must be " +
                                                "specified for every " +
                                                elDesc + " element.");

      wsdlExc.setLocation(XPathUtils.getXPathExprFromNode(el));

      throw wsdlExc;
    }
    else
    {
      return null;
    }
  }
  */

  public static void throwWSDLException(Element location) throws WSDLException
  {
    String elName = DOMQNameUtils.newQName(location).toString();

    WSDLException wsdlExc = new WSDLException(WSDLException.INVALID_WSDL,
                                              "Encountered unexpected element '" +
                                              elName + "'.");

    wsdlExc.setLocation(XPathUtils.getXPathExprFromNode(location));

    throw wsdlExc;
  }

  public static void throwWSDLException(Element location, List remainingAttrs) throws WSDLException
  {
    String elName = DOMQNameUtils.newQName(location).toString();
    
    StringBuffer sb = new StringBuffer();
    ListIterator i = remainingAttrs.listIterator();
    while (i.hasNext())
    {
      String attrName = DOMQNameUtils.newQName((Attr)i.next()).toString();
      sb.append(attrName);
      sb.append( i.hasNext() ? " " : "");
    }

    WSDLException wsdlExc = new WSDLException(WSDLException.INVALID_WSDL,
                                              "Element '" +
                                              elName +
                                              "' contained unexpected attributes: '" +
                                              sb.toString() +
                                              "'");

    wsdlExc.setLocation(XPathUtils.getXPathExprFromNode(location));

    throw wsdlExc;
  }

  /*
  public static void printAttribute(String name,
                                    String value,
                                    PrintWriter pw)
  {
    if (value != null)
    {
      pw.print(' ' + name + "=\"" + cleanString(value) + '\"');
    }
  }
  */

  /**
   * Prints attributes with qualified names.
   */
  /*
  public static void printQualifiedAttribute(QName name,
                                             String value,
                                             Definition def,
                                             PrintWriter pw)
                                               throws WSDLException
  {
    if (name != null)
    {
      printAttribute(getQualifiedValue(name.getNamespaceURI(),
                                       name.getLocalPart(),
                                       def),
                     value,
                     pw);
    }
  }

  public static void printQualifiedAttribute(QName name,
                                             QName value,
                                             Definition def,
                                             PrintWriter pw)
                                               throws WSDLException
  {
    if (value != null)
    {
      printAttribute(getQualifiedValue(name.getNamespaceURI(),
                                       name.getLocalPart(),
                                       def),
                     getQualifiedValue(value.getNamespaceURI(),
                                       value.getLocalPart(),
                                       def),
                     pw);
    }
  }

  public static void printQualifiedAttribute(String name,
                                             QName value,
                                             Definition def,
                                             PrintWriter pw)
                                               throws WSDLException
  {
    if (value != null)
    {
      printAttribute(name,
                     getQualifiedValue(value.getNamespaceURI(),
                                       value.getLocalPart(),
                                       def),
                     pw);
    }
  }

  public static String getQualifiedValue(String namespaceURI,
                                         String localPart,
                                         Definition def)
                                           throws WSDLException
  {
    String prefix = null;

    if (namespaceURI != null && !namespaceURI.equals(""))
    {
      prefix = getPrefix(namespaceURI, def);
    }

    return ((prefix != null && !prefix.equals(""))
            ? prefix + ":"
            : "") + localPart;
  }

  public static String getPrefix(String namespaceURI,
                                 Definition def)
                                   throws WSDLException
  {
    String prefix = def.getPrefix(namespaceURI);

    if (prefix == null)
    {
      throw new WSDLException(WSDLException.OTHER_ERROR,
                              "Can't find prefix for '" + namespaceURI +
                              "'. Namespace prefixes must be set on the" +
                              " Definition object using the " +
                              "addNamespace(...) method.");
    }

    return prefix;
  }

  public static String cleanString(String orig)
  {
    if (orig == null)
    {
      return "";
    }

    StringBuffer strBuf = new StringBuffer();
    char[] chars = orig.toCharArray();
    boolean inCDATA = false;

    for (int i = 0; i < chars.length; i++)
    {
      if (!inCDATA)
      {
        switch (chars[i])
        {
          case '&'  : strBuf.append("&amp;");
                      break;
          case '\"' : strBuf.append("&quot;");
                      break;
          case '\'' : strBuf.append("&apos;");
                      break;
          case '<'  :
                      {
                        if (chars.length >= i + 9)
                        {
                          String tempStr = new String(chars, i, 9);

                          if (tempStr.equals("<![CDATA["))
                          {
                            strBuf.append(tempStr);
                            i += 8;
                            inCDATA = true;
                          }
                          else
                          {
                            strBuf.append("&lt;");
                          }
                        }
                        else
                        {
                          strBuf.append("&lt;");
                        }
                      }
                      break;
          case '>'  : strBuf.append("&gt;");
                      break;
          default   : strBuf.append(chars[i]);
                      break;
        }
      }
      else
      {
        strBuf.append(chars[i]);

        if (chars[i] == '>'
            && chars[i - 1] == ']'
            && chars[i - 2] == ']')
        {
          inCDATA = false;
        }
      }
    }

    return strBuf.toString();
  }
  */
  
}

