/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
package org.apache.woden.wsdl20.xml;

import java.net.URI;
import java.util.Map;

import org.apache.woden.wsdl20.Description;
import org.apache.woden.wsdl20.extensions.ExtensionRegistry;

/**
 * Represents the WSDL 2.0 &lt;description&gt; element. 
 * 
 * @author John Kaputin (jkaputin@apache.org)
 */
public interface DescriptionElement extends DocumentableElement 
{

    /**
     * Set the document base URI to the specified URI. 
     * This is the base URI used to locate the &lt;description&gt; element that this 
     * DescriptionElement corresponds to.
     * This is used to resolve relative paths specified within this WSDL description that
     * refer to other WSDL or Schema documents. For example, via &lt;xs:import&gt;, &lt;wsdl:import&gt; 
     * or &lt;wsdl:include&gt; elements.
     * 
     * @param documentBaseURI the base URI of the WSDL
     */
    public void setDocumentBaseURI(URI documentBaseURI);
    
    /**
     * Return the document base URI that was used to locate the &lt;description&gt; element that 
     * this DescriptionElement corresponds to. This is used to resolve relative paths within this
     * WSDL description that refer to other WSDL or Schema documents.
     * 
     * @return the document base URI
     */
    public URI getDocumentBaseURI();
    
    /*
     * Attributes and namespaces
     */
    
    /**
     * Set the target namespace to the specified URI. 
     * This corresponds to the <code>targetNamespace</code> attribute of the &lt;description&gt;
     * element.
     * 
     * @param namespaceURI the target namespace URI
     */
    public void setTargetNamespace(URI namespaceURI);
    
    /**
     * Return the target namespace URI.
     * This corresponds to the <code>targetNamespace</code> attribute of the &lt;description&gt;
     * element.
     * 
     * @return the target namespace URI.
     */
    public URI getTargetNamespace();

    /**
     * Associate the specified prefix with the specified namespace URI.
     * This equates to adding an <code>xmlns</code> namespace declaration to the 
     * &lt;description&gt; element. 
     * To define the default namespace, specify null or the empty string "" for the prefix.
     * If null is specified for the namespace URI, the prefix/namespace association will be
     * removed (i.e. the same behaviour as the <code>removeNamespace</code> method).
     * If the specified prefix is already associated with a namespace URI, 
     * that association will be replaced by the specified prefix/namespace association. 
     *
     * @param prefix the prefix String associated with <code>namespace</code>
     * @param namespace the namespace URI associated with <code>prefix</code>
     */
    public void addNamespace(String prefix, URI namespace);
    
    /**
     * Remove the namespace URI associated with the specified prefix.
     * This equates to removing an <code>xmlns</code> namespace declaration from the 
     * &lt;description&gt; element.
     * To remove the default namespace, specify null or the empty string "" for the prefix.
     * If the specified prefix is not associated with a namespace, no action is performed.
     * 
     * @param prefix the prefix String associated with a namespace
     */
    public void removeNamespace(String prefix);
    
    /**
     * Return the namespace URI associated with the specified prefix.
     * This equates to an <code>xmlns</code> namespace declaration on the 
     * &lt;description&gt; element.
     * To return the default namespace, specify null or the empty string "" for the prefix.
     * If the specified prefix is not associated with a namespace, null is returned.
     * 
     * @param prefix the prefix String associated with a namespace
     */
    public URI getNamespace(String prefix);
    
    /**
     * Return the set of all prefix/namespace URI associations.
     * This equates to all of the <code>xmlns</code> namespace declaration on the 
     * &lt;description&gt; element.
     * 
     * @deprecated in M7, to be replaced in M8 with a type-safe return type (WODEN-140)
     * 
     * @return a Map of prefix String / namespace URI pairs 
     */
    public Map getNamespaces(); //TODO type-safe return type
    
    /*
     * Element factory methods
     */
    
    /**
     * Create an ImportElement with this DescriptionElement as its parent and
     * return a reference to it.
     * This equates to adding an &lt;import&gt; element to the &lt;description&gt; element.
     * 
     * @return the ImportElement
     */
    public ImportElement addImportElement();
    
    /**
     * Create an IncludeElement with this DescriptionElement as its parent and
     * return a reference to it.
     * This equates to adding an &lt;include&gt; element to the &lt;description&gt; element.
     * 
     * @return the IncludeElement
     */
    public IncludeElement addIncludeElement();

    /**
     * Create an InterfaceElement with this DescriptionElement as its parent and
     * return a reference to it.
     * This equates to adding an &lt;interface&gt; element to the &lt;description&gt; element.
     * 
     * @return the InterfaceElement
     */
    public InterfaceElement addInterfaceElement();

    /**
     * Create a BindingElement with this DescriptionElement as its parent and
     * return a reference to it.
     * This equates to adding an &lt;binding&gt; element to the &lt;description&gt; element.
     * 
     * @return the BindingElement
     */
    public BindingElement addBindingElement();

    /**
     * Create a ServiceElement with this DescriptionElement as its parent and
     * return a reference to it.
     * This equates to adding an &lt;service&gt; element to the &lt;description&gt; element.
     * 
     * @return the ServiceElement
     */
    public ServiceElement addServiceElement();
    
    /*
     * Element accessor and modifier methods
     * 
     * TODO removeXXX(obj), getXXX(key) methods
     * 
     */
    
    /**
     * Return the set of ImportElements within this DescriptionElement.
     * This equates to the set of &lt;import&gt; elements within the &lt;description&gt; element.
     * If no ImportElements exist, an empty array is returned.
     * 
     * @return an array of ImportElement
     */
    public ImportElement[] getImportElements();
    
    /**
     * Return the set of IncludeElements within this DescriptionElement.
     * This equates to the set of &lt;include&gt; elements within the &lt;description&gt; element.
     * If no IncludeElements exist, an empty array is returned.
     * 
     * @return an array of IncludeElement
     */
    public IncludeElement[] getIncludeElements();
    
    /**
     * Return the TypesElement within this DescriptionElement.
     * This corresponds to the &lt;types&gt; element within the &lt;description&gt; element.
     * If no TypesElement exists, one will be created with this DescriptionElement as its parent.
     * <p>
     * TODO modify this behaviour to return null if none exists and define a new factory method called createTypesElement() WODEN-141
     * 
     * @return the TypesElement
     */
    public TypesElement getTypesElement();
    
    /**
     * Return the set of InterfaceElements within this DescriptionElement.
     * This equates to the set of &lt;interface&gt; elements within the &lt;description&gt; element.
     * If no InterfaceElements exist, an empty array is returned.
     * 
     * @return an array of InterfaceElement
     */
    public InterfaceElement[] getInterfaceElements();
    
    /**
     * Return the set of BindingElements within this DescriptionElement.
     * This equates to the set of &lt;binding&gt; elements within the &lt;description&gt; element.
     * If no BindingElements exist, an empty array is returned.
     * 
     * @return an array of BindingElement
     */
    public BindingElement[] getBindingElements();
    
    /**
     * Return the set of ServiceElements within this DescriptionElement.
     * This equates to the set of &lt;service&gt; elements within the &lt;description&gt; element.
     * If no ServiceElements exist, an empty array is returned.
     * 
     * @return an array of ServiceElement
     */
    public ServiceElement[] getServiceElements();

    /**
     * Specify the ExtensionRegistry to be used handle WSDL extensions defined by this
     * DescriptionElement and by any WSDL elements nested within it.
     * 
     * @param extReg the ExtensionRegistry used with this description
     */
    public void setExtensionRegistry(ExtensionRegistry extReg);
    
    /**
     * Return the ExtensionRegistry used handle WSDL extensions defined by this
     * DescriptionElement and by any WSDL elements nested within it.
     * 
     * @return the ExtensionRegistry used with this description
     */
    public ExtensionRegistry getExtensionRegistry();

    /**
     * Return the Description component derived from this DescriptionElement.
     * <p>
     * The Description component is unlike other WSDL components which map neatly to a single 
     * WSDL element. Description represents a 'flattened' view of the entire WSDL infoset tree, 
     * including the WSDL components from any imported or included WSDL documents, so it could 
     * contain the content of multiple &lt;description&gt; elements.
     * <p>
     * If this DescriptionElement imports or includes other DescriptionElements (that is,
     * if it is the root of a composite WSDL document), the behaviour of this method is equivalent 
     * to traversing the WSDL tree recursively and aggregating the WSDL content of each 
     * DescriptionElement into a single Description component.
     * <p>
     * This equates to retrieving the root &lt;description&gt; element and traversing any
     * &lt;wsdl:import&gt; or &lt;wsdl:include&gt; elements to retrieve WSDL content of
     * any nested &lt;description&gt; elements. 
     * <p>
     * The <code>toComponent()</code> method is only defined for DescriptionElement, not for 
     * other WSDL Element interfaces. This is because the WSDL 2.0 Component model is a nested
     * hierarchy with a single Description component at the top. To access WSDL components,
     * we must start at the Description and traverse the component model from there.

     * @return the Description component derived from this DescriptionElement
     */
    public Description toComponent();
    
    
}
