/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
package org.apache.woden.wsdl20.xml;

import java.net.URI;

import javax.xml.namespace.QName;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.apache.woden.internal.ErrorReporterImpl;
import org.apache.woden.internal.schema.ImportedSchemaImpl;
import org.apache.woden.internal.wsdl20.Constants;
import org.apache.woden.internal.wsdl20.DescriptionImpl;
import org.apache.woden.internal.wsdl20.ElementDeclarationImpl;
import org.apache.woden.internal.wsdl20.InterfaceMessageReferenceImpl;
import org.apache.woden.internal.wsdl20.extensions.InterfaceOperationExtensionsImpl;
import org.apache.woden.schema.Schema;
import org.apache.woden.types.NCName;
import org.apache.woden.wsdl20.ElementDeclaration;
import org.apache.woden.wsdl20.InterfaceOperation;
import org.apache.woden.wsdl20.enumeration.Direction;
import org.apache.woden.wsdl20.extensions.ExtensionRegistry;
import org.apache.ws.commons.schema.XmlSchema;
import org.apache.ws.commons.schema.XmlSchemaCollection;
import org.apache.ws.commons.schema.XmlSchemaElement;



/**
 * Unit tests for the InterfaceMessageReferenceElement class.
 * 
 * @author Graham Turrell (gturrell@apache.org)
 */
public class InterfaceMessageReferenceElementTest extends TestCase {

	private InterfaceMessageReferenceElement fMessageReference = null;
	private ExtensionRegistry fExtensionRegistry = null;
	private URI fNamespace = null;

	public static Test suite()
	{
	   return new TestSuite(InterfaceMessageReferenceElementTest.class);
	   
	}
	   /*
     * @see TestCase#setUp()
     */
    protected void setUp() throws Exception 
    {
    	super.setUp();
    	fMessageReference = new InterfaceMessageReferenceImpl();
    	fNamespace = new URI("http://apache.org/testns");
    	fExtensionRegistry = new ExtensionRegistry();
        fExtensionRegistry.setErrorReporter(new ErrorReporterImpl());  	
    }
    
    /*
     * @see TestCase#tearDown()
     */
    protected void tearDown() throws Exception 
    {
        super.tearDown();
    }
	
	/*
	 * Test that a (mandatory) direction can be successfully set and retrieved
	 */
	public void testSetGetDirection()
	{
		// Default case
		assertNull("The retrieved Direction when unset should be null", fMessageReference.getDirection());
		
		fMessageReference.setDirection(Direction.OUT);
		assertEquals("The retrieved MessageReference direction is not that which was set", 
				Direction.OUT, fMessageReference.getDirection());
	}

	/*
	 * Test that the (Mandatory) message label attribute ("messageLabel") can be successfully set and retrieved
	 */
	public void testSetGetMessageLabel()
	{
		NCName messageRefNCName = new NCName("messageRefName");
		fMessageReference.setMessageLabel(messageRefNCName);
		assertEquals("The retrieved messageLabel is not that which was set", 
				messageRefNCName, fMessageReference.getMessageLabel());
	}

	/* 
	 * Test that the (Mandatory) Message Content Model property can be successfully set and retrieved 
	 */
	public void testSetGetMessageContentModel()
	{
		fMessageReference.setMessageContentModel("contentModel");
		assertEquals("The retrieved Message Content Model is not that which was set", 
				"contentModel", fMessageReference.getMessageContentModel());
	}
	

	/*
	 * Test that the (Mandatory) Element Name attribute ("element") can be successfully set and retrieved
	 */
	public void testSetGetElementName()
	{
		QName elementName = new QName("ElementName");
		fMessageReference.setElementName(elementName);
		assertEquals("The retrieved Element Name attribute is not that which was set", 
				elementName, fMessageReference.getElementName());
	}
	
	/* 
	 * Test that the (Optional) Element Declaration property can be successfully retrieved.
	 * The element reference is to an Element Declaration associated with a TypeDef below the ancestor Description,
	 * of the name set in setElementName().
	 * 
	 * TODO Need to check model structure for XmlSchema
	 */

	public void testGetElement()
	{
		// Create the DescriptionElement->InterfaceElement->InterfaceOperationElement->InterfaceMessageReferenceElement hierarchy
		DescriptionElement descriptionElement = new DescriptionImpl();
		InterfaceElement interfaceElement = descriptionElement.addInterfaceElement();
		InterfaceOperationElement interfaceOperationElement = interfaceElement.addInterfaceOperationElement();
		
		// TODO - should this be strictly necessary for DescriptionElement.toComponent() not to NPE?
        descriptionElement.setExtensionRegistry(fExtensionRegistry);
		
		// create the InterfaceMessageReferenceElement to test
		InterfaceMessageReferenceElement messageReference = interfaceOperationElement.addInterfaceMessageReferenceElement();
		
		// Default case:
		XmlSchemaElement retrievedElement = messageReference.getElement();
		assertNull("Unset Element Declaration should return null", retrievedElement);

		// populated case 1 - (with Message Content Model set to #ANY or #NONE).
		ElementDeclaration ed = new ElementDeclarationImpl();
		((DescriptionImpl)descriptionElement).addElementDeclaration(ed);		
		messageReference.setMessageContentModel(Constants.NMTOKEN_NONE);

		descriptionElement.toComponent(); 

		retrievedElement = messageReference.getElement();
		assertNull("The retrieved Element Declation should return null when message content model is #NONE", 
				retrievedElement);
		
		messageReference.setMessageContentModel(Constants.NMTOKEN_ANY);		
		descriptionElement.toComponent(); 
		retrievedElement = messageReference.getElement();
		assertNull("The retrieved Element Declaration should return null when message content model is #ANY", 
				retrievedElement);
		
		/* TODO following populated case (with Message Content Model set) requires XmlSchema "element" field to be set.
		 * (probably via org.apache.ws.commons.schema.SchemaBuilder.handleXmlSchemaElement())
		 */
/*      
		// populated case 2 - success:
				
		// Add a TypesElement to the Description
		TypesElement typesElement = descriptionElement.getTypesElement();
		
		// Add a namespace-identified schema element declaration to the Types element
		XmlSchemaCollection xsc = new XmlSchemaCollection();
		XmlSchema xmlSchema = new XmlSchema("ElementName", xsc);
		Schema schema = new ImportedSchemaImpl(); // arbitrary schema
		schema.setNamespace(fNamespace);
		schema.setSchemaDefinition(xmlSchema);
		typesElement.addSchema(schema);
				        
        fExtensionRegistry.registerComponentExtension(InterfaceOperation.class, 
        		InterfaceOperationExtensionsImpl.URI_NS_EXTENSIONS, InterfaceOperationExtensionsImpl.class);
    
	    descriptionElement.setExtensionRegistry(fExtensionRegistry);
		
		// set the element name reference
	    QName elementName = new QName("http://apache.org/testns", "ElementName");
		messageReference.setElementName(elementName);
		messageReference.setMessageContentModel(Constants.NMTOKEN_ELEMENT);
		
		//descriptionElement.toComponent(); // quick test of model integrity for testing purposes
		
		retrievedElement = messageReference.getElement();
		assertEquals("The retrieved Element is not that which was set", 
				schema, retrievedElement);
*/
	}
}
